import unittest
import math

from . import eval_expr
from ._grammar import _symbol_table


class ExprTest(unittest.TestCase):

    def test_add(self):
        expr = "3 + 4"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_sub(self):
        expr = "6 - 2"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_div(self):
        expr = "9.32 / 2"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_mul(self):
        expr = "3.1415 * 7.2"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_expr(self):
        expr = "-3.21^5.6"
        expr_py = "-3.21**5.6"
        self.assertEqual(eval_expr(expr), eval(expr_py))

    def test_float(self):
        expr = "31.415"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_paren(self):
        expr = "(3 + 4 - 2 + 6.2) / 2"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_one_arg_func(self):
        expr = "sin(3 + 4 / 2)"
        expr_py = expr.replace("sin", "math.sin")
        self.assertEqual(eval_expr(expr), eval(expr_py))

    def test_mult_arg_func(self):
        expr = "max(3 + 4 / 4, 7 + 3)"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_var(self):
        nice_var_name = 2.32
        _symbol_table["nice_var_name"] = nice_var_name
        expr = "3 + nice_var_name / 4"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_standalone_var(self):
        x = 2.32
        _symbol_table["x"] = 2.32
        expr = "x"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_neg_exp(self):
        x = 5.32
        _symbol_table["x"] = 5.32
        expr = "-x^4.90"
        expr_py = "-x ** 4.90"
        self.assertEqual(eval_expr(expr), eval(expr_py))

    def test_neg(self):
        expr = "-3.294 + -9.2"
        self.assertEqual(eval_expr(expr), eval(expr))

    def test_neg_paren(self):
        expr = "-(3 - 9) / 10"
        self.assertEqual(eval_expr(expr), eval(expr))


class ErrorTest(unittest.TestCase):
    pass



from ._grammar import _operators, _LETTERS, _NUMBERS, _FUNCTION_MAP, _operators_grouping, _precedence


def _peek(stack: list):
    if stack:
        return stack[len(stack)-1]


def _inverse(char: str):
    return "(" if char == ")" or char == "," else "["


def _clean(expression: str):
    i = 0
    expression = list(expression)

    current_token = None
    previous_token = None

    while i < len(expression):
        if expression[i] == " " or expression[i] == "\t" or expression[i] == "\n":
            i += 1
            continue
        current_token = expression[i]

        if expression[i] in _operators:
            if previous_token is None:
                if expression[i] == "+":
                    expression[i] = ">"
                elif expression[i] == "-":
                    expression[i] = "<"
            elif previous_token and not (previous_token.isdigit() or previous_token.upper() in (_LETTERS + "_")) \
                    and not previous_token in [")", "]"]:
                if expression[i] == "+":
                    expression[i] = ">"
                elif expression[i] == "-":
                    expression[i] = "<"

        previous_token = current_token
        i += 1

    return "".join(expression)


def _build_number(expression: str, pos: int):
    result = []
    dot_count = 0

    while pos < len(expression) and (expression[pos].isdigit() or
                                     expression[pos] == "." and dot_count < 1):

        if expression[pos] == ".":
            dot_count += 1
        result.append(expression[pos])
        pos += 1

    final = "".join(result)
    if final == ".":
        raise ValueError("'.' is not a valid number.")

    return final, pos


def _build_identifier(expression: str, pos: int):
    result = []
    while pos < len(expression) and expression[pos].upper() in (_LETTERS + _NUMBERS + "_"):
        result.append(expression[pos])
        pos += 1

    final = "".join(result)
    _type = "func" if final in _FUNCTION_MAP else "id"
    return final, pos, _type


def _infix_to_postfix(expression: str):
    if not _is_balanced(expression):
        raise SyntaxError("Missing grouping symbol")
    expression = _clean(expression)

    output = []
    stack = []
    i = 0

    while i < len(expression):
        if expression[i].isdigit() or expression[i] == ".":
            result, i = _build_number(expression, i)
            output.append(result + " ")
            continue
        elif expression[i].upper() in _LETTERS:
            result, i, _type = _build_identifier(expression, i)
            if _type == "id":
                output.append(result + " ")
            else:
                stack.append(result)
            continue
        elif expression[i] in _operators_grouping:
            right_associative = expression[i] == "^" or expression[i] == ">" or expression[i] == "<"
            if _precedence[expression[i]] == 6 or expression[i] == ",":
                while _peek(stack) != _inverse(expression[i]):
                    if len(stack) == 0:
                        raise SyntaxError("Missing '('")
                    output.append(stack.pop())

                if expression[i] != ",":
                    if len(stack) == 0:
                        raise SyntaxError("Missing '('")
                    stack.pop()
                if _peek(stack) in _FUNCTION_MAP:
                    output.append(stack.pop() + " ")
            elif len(stack) == 0 or _precedence[_peek(stack)] < _precedence[expression[i]] \
                    or _precedence[expression[i]] == 0:
                stack.append(expression[i])
            else:
                while len(stack) > 0 and _precedence[_peek(stack)] >= _precedence[expression[i]] \
                        and not right_associative:
                    output.append(stack.pop())
                stack.append(expression[i])
        i += 1

    while len(stack) > 0:
        output.append(stack.pop())

    return "".join(output)


def _is_balanced(expression: str):
    open_group_symbols = ["[", "("]
    close_group_symbols = ["]", ")"]

    stack = []

    for i in range(len(expression)):
        if expression[i] not in open_group_symbols and expression[i] not in close_group_symbols:
            continue

        if expression[i] in open_group_symbols:
            stack.append(expression[i])
        elif expression[i] in close_group_symbols:
            if (_peek(stack) == _inverse(expression[i])):
                stack.pop()
            else:
                return False

    return len(stack) == 0


if __name__ == "__main__":
    print(_is_balanced("(3 + 4"))



from . import _shunting_yard as shunt
from . import _postfix_parser as postfix_parser
from . import _grammar as grammar


class TPExpr:
    def __init__(self, expr: str):
        self.__expr = expr

    def get_expr(self):
        return self.__expr

    def __repr__(self):
        return self.__expr

    def __str__(self):
        return self.__expr


def eval_expr(expression: str):
    return shunt._eval_postfix(postfix_parser._infix_to_postfix(expression))


def compile_expr(expression: str):
    return TPExpr(postfix_parser._infix_to_postfix(expression))


def eval_compiled(expression: TPExpr):
    return shunt._eval_postfix(expression.get_expr())


def add_var(var_name: str, var: float):
    grammar._symbol_table[var_name] = var


def update_var(var_name: str, var: float):
    grammar._symbol_table[var_name] = var


def add_func(func_name: str, func):
    grammar._FUNCTION_MAP[func_name] = func
    grammar._precedence[func_name] = 1


def update_func(func_name: str, func):
    grammar._FUNCTION_MAP[func_name] = func



from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

VERSION = "0.0.2"
DESCRIPTION = "An infix-to-postfix expression parser for evaluating math expressions."

setup(
    name="tinypost",
    version=VERSION,
    author="Nathan Abraham",
    author_email="abnathan123@gmail.com",
    description=DESCRIPTION,
    packages=find_packages(),
    keywords=["python", "math", "shunting-yard",
              "parser", "expression evaluation"],
    classifiers=["Operating System :: Unix",
                 "Operating System :: MacOS :: MacOS X",
                 "Operating System :: Microsoft :: Windows",
                 ]
)



import math
from ._functions import _log, _max_two

_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
_NUMBERS = "0123456789"
_operators = ["+", "-", "*", "/", "^", ">", "<", ",", "%"]
_operators_grouping = _operators + ["(", "[", ")", "]"]

_precedence = {
    ")": 6,
    "]": 6,
    ",": 6,

    "^": 4,

    "<": 4,
    ">": 4,

    "*": 3,
    "/": 3,
    "%": 3,

    "+": 2,
    "-": 2,

    "(": 0,
    "[": 0,

    None: float("inf"),
}

_BUILTIN_FUNCTION_MAP = {
    "abs": abs,
    "sqrt": math.sqrt,
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "ln": _log,
    "acos": math.acos,
    "asin": math.asin,
    "atan": math.atan,
    "exp": math.exp,
    "log": math.log10,
    "sinh": math.sinh,
    "cosh": math.cosh,
    "tanh": math.tanh,

    "atan2": math.atan2,
    "max": _max_two,
    "pow": math.pow,
}

_FUNCTION_MAP = {
    **_BUILTIN_FUNCTION_MAP,
}

for function in _FUNCTION_MAP:
    _precedence[function] = 1

_symbol_table = {
    "nice_var_name": 10,
    "x": 5,
    "pi": math.pi,
    "e": math.e,
}

if __name__ == "__main__":
    print(_FUNCTION_MAP)



import math
import time
from inspect import signature

from ._postfix_parser import _build_number, _build_identifier, _infix_to_postfix, _clean
from ._grammar import _operators, _LETTERS, _FUNCTION_MAP, _symbol_table

error = None


def _eval_func(func_name, params):
    return _FUNCTION_MAP[func_name](*params)


def _eval_basic(left_operand, right_operand, operator):
    left = float(left_operand)
    right = float(right_operand)

    if operator == "+" or operator == ">":
        return left + right
    elif operator == "-" or operator == "<":
        return left - right
    elif operator == "*":
        return left * right
    elif operator == "/":
        return left / right
    elif operator == "^":
        return left ** right
    elif operator == "%":
        return left % right


def _eval_postfix(expression: str):
    global error, _symbol_table

    stack = []
    i = 0
    error = None

    current_token = None

    while i < len(expression):
        current_token = expression[i]
        if expression[i].isdigit() or expression[i] == ".":
            current_token, i = _build_number(expression, i)
            stack.append(current_token)
            continue
        elif expression[i].upper() in _LETTERS:
            current_token, i, _type = _build_identifier(expression, i)
            if _type == "id":
                if current_token in _symbol_table:
                    stack.append(str(_symbol_table[current_token]))
                else:
                    return "Variable not found"
                continue
        if current_token in _FUNCTION_MAP or current_token in _operators:
            if expression[i] == ">" or expression[i] == "<":
                if len(stack) == 0:
                    raise SyntaxError("Missing operand after unary operator")

                temp = stack.pop()
                stack.append("0")
                stack.append(temp)
            if len(stack) < 2 and current_token in _operators:
                error = i
                raise SyntaxError("Not enough operands")
                # return f"Syntax error at position {i}"
            else:
                if current_token in _FUNCTION_MAP:
                    num_params = len(
                        signature(_FUNCTION_MAP[current_token]).parameters)
                    params = []
                    for _ in range(num_params):
                        if len(stack) == 0:
                            raise SyntaxError("Missing function arguments")

                        params.insert(0, float(stack.pop()))
                    temp_result = str(_eval_func(current_token, params))
                else:
                    if len(stack) == 0:
                        raise SyntaxError("Missing right operand")
                    right_operand = stack.pop()

                    if len(stack) == 0:
                        raise SyntaxError("Missing left operand")
                    left_operand = stack.pop()

                    current_token = expression[i]
                    temp_result = str(_eval_basic(
                        left_operand, right_operand, current_token))

                stack.append(temp_result)

        i += 1

    if len(stack) != 1:
        return "Invalid Expression"
    else:
        return float(stack[0])


if __name__ == "__main__":
    expr = "+-3"
    expr2 = "3*2+4*5"

    print(_clean(expr))
    print(_infix_to_postfix(expr))

    start = time.time()
    print(_eval_postfix(_infix_to_postfix(expr)))
    end = time.time()
    print(f"Took {end-start} seconds")

    # print(clean(expr2))
    # print(infix_to_postfix(expr2))
    # print(eval_expr(expr2))
